# Read-only methods are commented out

class SerializedList(list):
    # def __add__(self, *args, **kwargs):
    #     super().__add__(*args, **kwargs)
    #     self.checked_serialize()

    # def __class__(self, *args, **kwargs):
    #     super().__class__(*args, **kwargs)
    #     self.checked_serialize()

    # def __contains__(self, *args, **kwargs):
    #     super().__contains__(*args, **kwargs)
    #     self.checked_serialize()

    # def __delattr__(self, *args, **kwargs):
    #     super().__delattr__(*args, **kwargs)
    #     self.checked_serialize()

    def __delitem__(self, *args, **kwargs):
        r = super().__delitem__(*args, **kwargs)
        self.checked_serialize()
        return r

    # def __dir__(self, *args, **kwargs):
    #     super().__dir__(*args, **kwargs)
    #     self.checked_serialize()

    # def __doc__(self, *args, **kwargs):
    #     super().__doc__(*args, **kwargs)
    #     self.checked_serialize()

    # def __eq__(self, *args, **kwargs):
    #     super().__eq__(*args, **kwargs)
    #     self.checked_serialize()

    # def __format__(self, *args, **kwargs):
    #     super().__format__(*args, **kwargs)
    #     self.checked_serialize()

    # def __ge__(self, *args, **kwargs):
    #     super().__ge__(*args, **kwargs)
    #     self.checked_serialize()

    # def __getattribute__(self, *args, **kwargs):
    #     super().__getattribute__(*args, **kwargs)
    #     self.checked_serialize()

    # def __getitem__(self, *args, **kwargs):
    #     super().__getitem__(*args, **kwargs)
    #     self.checked_serialize()

    # def __gt__(self, *args, **kwargs):
    #     super().__gt__(*args, **kwargs)
    #     self.checked_serialize()

    # def __hash__(self, *args, **kwargs):
    #     super().__hash__(*args, **kwargs)
    #     self.checked_serialize()

    def __iadd__(self, *args, **kwargs):
        r = super().__iadd__(*args, **kwargs)
        self.checked_serialize()
        return r

    def __imul__(self, *args, **kwargs):
        r = super().__imul__(*args, **kwargs)
        self.checked_serialize()
        return r

    # def __init__(self, *args, **kwargs):
    #     super().__init__(*args, **kwargs)
    #     self.checked_serialize()

    # def __init_subclass__(self, *args, **kwargs):
    #     super().__init_subclass__(*args, **kwargs)
    #     self.checked_serialize()

    # def __iter__(self, *args, **kwargs):
    #     super().__iter__(*args, **kwargs)
    #     self.checked_serialize()

    # def __le__(self, *args, **kwargs):
    #     super().__le__(*args, **kwargs)
    #     self.checked_serialize()

    # def __len__(self, *args, **kwargs):
    #     super().__len__(*args, **kwargs)
    #     self.checked_serialize()

    # def __lt__(self, *args, **kwargs):
    #     super().__lt__(*args, **kwargs)
    #     self.checked_serialize()

    def __mul__(self, *args, **kwargs):
        r = super().__mul__(*args, **kwargs)
        self.checked_serialize()
        return r

    # def __ne__(self, *args, **kwargs):
    #     super().__ne__(*args, **kwargs)
    #     self.checked_serialize()

    # def __new__(self, *args, **kwargs):
    #     super().__new__(*args, **kwargs)
    #     self.checked_serialize()

    # def __reduce__(self, *args, **kwargs):
    #     super().__reduce__(*args, **kwargs)
    #     self.checked_serialize()

    # def __reduce_ex__(self, *args, **kwargs):
    #     super().__reduce_ex__(*args, **kwargs)
    #     self.checked_serialize()

    # def __repr__(self, *args, **kwargs):
    #     super().__repr__(*args, **kwargs)
    #     self.checked_serialize()

    # def __reversed__(self, *args, **kwargs):
    #     super().__reversed__(*args, **kwargs)
    #     self.checked_serialize()

    def __rmul__(self, *args, **kwargs):
        r = super().__rmul__(*args, **kwargs)
        self.checked_serialize()
        return r

    # def __setattr__(self, *args, **kwargs):
    #     super().__setattr__(*args, **kwargs)
    #     self.checked_serialize()

    def __setitem__(self, *args, **kwargs):
        r = super().__setitem__(*args, **kwargs)
        self.checked_serialize()
        return r

    # def __sizeof__(self, *args, **kwargs):
    #     super().__sizeof__(*args, **kwargs)
    #     self.checked_serialize()

    # def __str__(self, *args, **kwargs):
    #     super().__str__(*args, **kwargs)
        # self.checked_serialize()

    # def __subclasshook__(self, *args, **kwargs):
    #     super().__subclasshook__(*args, **kwargs)
    #     self.checked_serialize()

    def append(self, *args, **kwargs):
        r = super().append(*args, **kwargs)
        self.checked_serialize()
        return r

    def clear(self, *args, **kwargs):
        r = super().clear(*args, **kwargs)
        self.checked_serialize()
        return r

    # def copy(self, *args, **kwargs):
    #     super().copy(*args, **kwargs)
    #     self.checked_serialize()

    def count(self, *args, **kwargs):
        r = super().count(*args, **kwargs)
        self.checked_serialize()
        return r

    def extend(self, *args, **kwargs):
        r = super().extend(*args, **kwargs)
        self.checked_serialize()
        return r

    def index(self, *args, **kwargs):
        r = super().index(*args, **kwargs)
        self.checked_serialize()
        return r

    def insert(self, *args, **kwargs):
        r = super().insert(*args, **kwargs)
        self.checked_serialize()
        return r

    def pop(self, *args, **kwargs):
        r = super().pop(*args, **kwargs)
        self.checked_serialize()
        return r

    def remove(self, *args, **kwargs):
        r = super().remove(*args, **kwargs)
        self.checked_serialize()
        return r

    def reverse(self, *args, **kwargs):
        r = super().reverse(*args, **kwargs)
        self.checked_serialize()
        return r

    def sort(self, *args, **kwargs):
        r = super().sort(*args, **kwargs)
        self.checked_serialize()
        return r



# Read-only methods are commented out

class SerializedDict(dict):
    # def __class__(self, *args, **kwargs):
    #     r = super().__class__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __contains__(self, *args, **kwargs):
    #     r = super().__contains__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    def __delattr__(self, *args, **kwargs):
        r = super().__delattr__(*args, **kwargs)
        self.checked_serialize()
        return r

    def __delitem__(self, *args, **kwargs):
        r = super().__delitem__(*args, **kwargs)
        self.checked_serialize()
        return r

    # def __dir__(self, *args, **kwargs):
    #     r = super().__dir__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __doc__(self, *args, **kwargs):
    #     r = super().__doc__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __eq__(self, *args, **kwargs):
    #     r = super().__eq__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __format__(self, *args, **kwargs):
    #     r = super().__format__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __ge__(self, *args, **kwargs):
    #     r = super().__ge__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __getattribute__(self, *args, **kwargs):
    #     r = super().__getattribute__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __getitem__(self, *args, **kwargs):
    #     r = super().__getitem__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __gt__(self, *args, **kwargs):
    #     r = super().__gt__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __hash__(self, *args, **kwargs):
    #     r = super().__hash__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __init__(self, *args, **kwargs):
    #     r = super().__init__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __init_subclass__(self, *args, **kwargs):
    #     r = super().__init_subclass__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __iter__(self, *args, **kwargs):
    #     r = super().__iter__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __le__(self, *args, **kwargs):
    #     r = super().__le__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __len__(self, *args, **kwargs):
    #     r = super().__len__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __lt__(self, *args, **kwargs):
    #     r = super().__lt__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __ne__(self, *args, **kwargs):
    #     r = super().__ne__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __new__(self, *args, **kwargs):
    #     r = super().__new__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __reduce__(self, *args, **kwargs):
    #     r = super().__reduce__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __reduce_ex__(self, *args, **kwargs):
    #     r = super().__reduce_ex__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __repr__(self, *args, **kwargs):
    #     r = super().__repr__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __reversed__(self, *args, **kwargs):
    #     r = super().__reversed__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __setattr__(self, *args, **kwargs):
    #     r = super().__setattr__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    def __setitem__(self, *args, **kwargs):
        r = super().__setitem__(*args, **kwargs)
        self.checked_serialize()
        return r

    # def __sizeof__(self, *args, **kwargs):
    #     r = super().__sizeof__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __str__(self, *args, **kwargs):
    #     r = super().__str__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def __subclasshook__(self, *args, **kwargs):
    #     r = super().__subclasshook__(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    def clear(self, *args, **kwargs):
        r = super().clear(*args, **kwargs)
        self.checked_serialize()
        return r

    # def copy(self, *args, **kwargs):
    #     r = super().copy(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def fromkeys(self, *args, **kwargs):
    #     r = super().fromkeys(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def get(self, *args, **kwargs):
    #     r = super().get(*args, **kwargs)
    #     self.checked_serialize()
    #     return r

    # def items(self, *args, **kwargs):
    #     r = super().items(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    # def keys(self, *args, **kwargs):
    #     r = super().keys(*args, **kwargs)
    #     self.checked_serialize()
    # return r

    def pop(self, *args, **kwargs):
        r = super().pop(*args, **kwargs)
        self.checked_serialize()
        return r

    def popitem(self, *args, **kwargs):
        r = super().popitem(*args, **kwargs)
        self.checked_serialize()
        return r

    def setdefault(self, *args, **kwargs):
        r = super().setdefault(*args, **kwargs)
        self.checked_serialize()
        return r

    def update(self, *args, **kwargs):
        r = super().update(*args, **kwargs)
        self.checked_serialize()
        return r

    # def values(self, *args, **kwargs):
    #     r = super().values(*args, **kwargs)
    #     self.checked_serialize()
    # return r



import warnings
import dis
import pickle
import json
import os
import atexit
import inspect

from typing import Dict, Any

from pickledict.dict import SerializedDict
from pickledict.list import SerializedList


def frames_up(n):
    # Offset by one to include this function call
    i = -1

    frame = inspect.currentframe()
    while i < n:
        frame = frame.f_back
        i += 1

    return frame


name_index = 0
seen_names = set()


class Loader:
    def __init__(self, save_on_every_write: bool = False, file_name: str = None, serializer_args: Dict[str, Any] = None, *args, **kwargs):
        if not os.path.exists(".pickledict"):
            os.mkdir(".pickledict")
        if file_name is None:
            file_name = os.path.join(".pickledict", self.get_name())
            if file_name in seen_names:
                warnings.warn(
                    f"The auto-generated file name '{file_name}' has already been used "
                    f"for this {self.__class__.__name__}"
                    f", please use {self.__class__.__name__}'s 'file_name' argument or you will get"
                    " incorrect data", stacklevel=2)
            seen_names.add(file_name)
        self.file_name = file_name

        self.save_on_every_write = save_on_every_write
        if serializer_args is None:
            serializer_args = {}
        self.serializer_args = serializer_args
        self.detached = False

        if not self.save_on_every_write:
            def exit_save():

                self.serialize()

            atexit.register(exit_save)

        if os.path.exists(file_name):
            try:
                deserialized = self.deserialize()
                super().__init__(deserialized)
            except:
                super().__init__(*args, **kwargs)
                self.serialize()
        else:
            super().__init__(*args, **kwargs)
            self.serialize()

    def detach(self):
        self.detached = True

    def generate_name(self, extension):
        try:
            f = frames_up(3)
            remaining_instructions = [x for x in dis.get_instructions(
                f.f_code) if x.offset > f.f_lasti]
            if len(remaining_instructions) > 0:
                maybe_store = remaining_instructions[0]
                if maybe_store.opname.startswith("STORE_"):
                    dest_var = maybe_store.argval
                    file_name = os.path.basename(f.f_code.co_filename)
                    if file_name.endswith(".py"):
                        file_name = file_name[:-len(".py")]

                    return f"{file_name}-{f.f_lineno}-{dest_var}.{extension}"
        except Exception as e:
            pass

        global name_index
        name_index += 1
        warnings.warn(
            f"Could not determine a name for this {self.__class__.__name__}"
            f", please use the 'file_name' argument or any changes"
            " to the order of pickledicts in your program will cause"
            " corrupted data", stacklevel=4)
        return f"{self.__class__.__name__}-{name_index}.{extension}"

    def checked_serialize(self):
        if not getattr(self, "detached", False) and \
                getattr(self, "save_on_every_write", False):
            self.serialize()


class PickleSerializer:
    def get_name(self) -> str:
        return Loader.generate_name(self, "pkl")

    def serialize(self) -> None:
        with open(self.file_name, "wb") as f:
            pickle.dump(self, f, **getattr(self, 'serializer_args', {}))

    def deserialize(self) -> Dict[Any, Any]:
        with open(self.file_name, "rb") as f:
            return pickle.load(f)


class JsonSerializer:
    def get_name(self) -> str:
        return Loader.generate_name(self, "json")

    def serialize(self) -> None:
        with open(self.file_name, "w") as f:
            json.dump(self, f, **getattr(self, 'serializer_args', {}))

    def deserialize(self) -> Dict[Any, Any]:
        with open(self.file_name, "r") as f:
            return json.load(f)


class pickledict(Loader, PickleSerializer, SerializedDict):
    pass


class jsondict(Loader, JsonSerializer, SerializedDict):
    pass


class picklelist(Loader, PickleSerializer, SerializedList):
    pass


class jsonlist(Loader, JsonSerializer, SerializedList):
    pass
